@@ -1,5 +1,6 @@
 # coding: utf-8
 
+
 class StixEntity:
     def __init__(self, opencti):
         self.opencti = opencti
@@ -23,6 +24,9 @@ class StixEntity:
                     description
                     created
                     modified
+                    ... on Organization {
+                        organization_class
+                    }
                 }
                 relation {
                     id
@@ -127,6 +131,9 @@ class StixEntity:
                 first_seen
                 last_seen
             }
+            ... on Organization {
+                organization_class
+            }
             ... on Malware {
                 killChainPhases {
                     edges {
@@ -146,6 +153,36 @@ class StixEntity:
                     }
                 }                
             }
+            ... on Tool {
+                tool_version
+                killChainPhases {
+                    edges {
+                        node {
+                            id
+                            entity_type
+                            stix_id_key
+                            kill_chain_name
+                            phase_name
+                            phase_order
+                            created
+                            modified
+                        }
+                        relation {
+                            id
+                        }
+                    }
+                }
+            }
+            ... on Vulnerability {
+                base_score
+                base_severity
+                attack_vector
+                integrity_impact
+                availability_impact
+            }            
+            ... on Organization {
+                organization_class
+            }
             ... on Indicator {
                 indicator_pattern
                 pattern_type
@@ -155,6 +192,24 @@ class StixEntity:
                             id
                             stix_id_key
                             entity_type
+                            observable_value
+                        }
+                        relation {
+                            id
+                        }
+                    }
+                }
+                killChainPhases {
+                    edges {
+                        node {
+                            id
+                            entity_type
+                            stix_id_key
+                            kill_chain_name
+                            phase_name
+                            phase_order
+                            created
+                            modified
                         }
                         relation {
                             id
@@ -185,6 +240,7 @@ class StixEntity:
                             id
                             stix_id_key
                             entity_type
+                            observable_value
                         }
                         relation {
                             id
@@ -204,6 +260,8 @@ class StixEntity:
                 }
             }            
             ... on StixRelation {
+                first_seen
+                last_seen
                 killChainPhases {
                     edges {
                         node {
@@ -232,20 +290,29 @@ class StixEntity:
     """
 
     def read(self, **kwargs):
-        id = kwargs.get('id', None)
+        id = kwargs.get("id", None)
+        custom_attributes = kwargs.get("customAttributes", None)
         if id is not None:
-            self.opencti.log('info', 'Reading Stix-Entity {' + id + '}.')
-            query = """
+            self.opencti.log("info", "Reading Stix-Entity {" + id + "}.")
+            query = (
+                """
                 query StixEntity($id: String!) {
                     stixEntity(id: $id) {
-                        """ + self.properties + """
+                        """
+                + (
+                    custom_attributes
+                    if custom_attributes is not None
+                    else self.properties
+                )
+                + """
                     }
                 }
              """
-            result = self.opencti.query(query, {'id': id})
-            return self.opencti.process_multiple_fields(result['data']['stixEntity'])
+            )
+            result = self.opencti.query(query, {"id": id})
+            return self.opencti.process_multiple_fields(result["data"]["stixEntity"])
         else:
-            self.opencti.log('error', 'Missing parameters: id')
+            self.opencti.log("error", "Missing parameters: id")
             return None
 
     """
@@ -257,28 +324,53 @@ class StixEntity:
     """
 
     def update_created_by_ref(self, **kwargs):
-        id = kwargs.get('id', None)
-        stix_entity = kwargs.get('entity', None)
-        identity_id = kwargs.get('identity_id', None)
+        id = kwargs.get("id", None)
+        identity_id = kwargs.get("identity_id", None)
         if id is not None and identity_id is not None:
+            custom_attributes = """
+                id
+                createdByRef {
+                    node {
+                        id
+                        entity_type
+                        stix_id_key
+                        stix_label
+                        name
+                        alias
+                        description
+                        created
+                        modified
+                        ... on Organization {
+                            organization_class
+                        }
+                    }
+                    relation {
+                        id
+                    }
+                }    
+            """
+            stix_entity = self.read(id=id, customAttributes=custom_attributes)
             if stix_entity is None:
-                stix_entity = self.read(id=id)
-            if stix_entity is None:
-                self.opencti.log('error', 'Cannot update created_by_ref, entity not found')
+                self.opencti.log(
+                    "error", "Cannot update created_by_ref, entity not found"
+                )
                 return False
             current_identity_id = None
             current_relation_id = None
-            if stix_entity['createdByRef'] is not None:
-                current_identity_id = stix_entity['createdByRef']['id']
-                current_relation_id = stix_entity['createdByRef']['remote_relation_id']
-
+            if stix_entity["createdByRef"] is not None:
+                current_identity_id = stix_entity["createdByRef"]["id"]
+                current_relation_id = stix_entity["createdByRef"]["remote_relation_id"]
             # Current identity is the same
             if current_identity_id == identity_id:
                 return True
             else:
                 self.opencti.log(
-                    'info',
-                    'Updating author of Stix-Entity {' + id + '} with Identity {' + identity_id + '}'
+                    "info",
+                    "Updating author of Stix-Entity {"
+                    + id
+                    + "} with Identity {"
+                    + identity_id
+                    + "}",
                 )
                 # Current identity is different, delete the old relation
                 if current_relation_id is not None:
@@ -291,7 +383,9 @@ class StixEntity:
                             }
                         }
                     """
-                    self.opencti.query(query, {'id': id, 'relationId': current_relation_id})
+                    self.opencti.query(
+                        query, {"id": id, "relationId": current_relation_id}
+                    )
                 # Add the new relation
                 query = """
                    mutation StixEntityEdit($id: ID!, $input: RelationAddInput) {
@@ -303,18 +397,18 @@ class StixEntity:
                    }
                 """
                 variables = {
-                    'id': id,
-                    'input': {
-                        'fromRole': 'so',
-                        'toId': identity_id,
-                        'toRole': 'creator',
-                        'through': 'created_by_ref'
-                    }
+                    "id": id,
+                    "input": {
+                        "fromRole": "so",
+                        "toId": identity_id,
+                        "toRole": "creator",
+                        "through": "created_by_ref",
+                    },
                 }
                 self.opencti.query(query, variables)
 
         else:
-            self.opencti.log('error', 'Missing parameters: id and identity_id')
+            self.opencti.log("error", "Missing parameters: id and identity_id")
             return False
 
     """
@@ -326,24 +420,46 @@ class StixEntity:
     """
 
     def add_marking_definition(self, **kwargs):
-        id = kwargs.get('id', None)
-        stix_entity = kwargs.get('entity', None)
-        marking_definition_id = kwargs.get('marking_definition_id', None)
+        id = kwargs.get("id", None)
+        marking_definition_id = kwargs.get("marking_definition_id", None)
         if id is not None and marking_definition_id is not None:
+            custom_attributes = """
+                id
+                markingDefinitions {
+                    edges {
+                        node {
+                            id
+                            entity_type
+                            stix_id_key
+                            definition_type
+                            definition
+                            level
+                            color
+                            created
+                            modified
+                        }
+                        relation {
+                            id
+                        }
+                    }
+                }
+            """
+            stix_entity = self.read(id=id, customAttributes=custom_attributes)
             if stix_entity is None:
-                stix_entity = self.read(id=id)
-            if stix_entity is None:
-                self.opencti.log('error', 'Cannot add Marking-Definition, entity not found')
+                self.opencti.log(
+                    "error", "Cannot add Marking-Definition, entity not found"
+                )
                 return False
-            markings_ids = []
-            for marking in stix_entity['markingDefinitions']:
-                markings_ids.append(marking['id'])
-            if marking_definition_id in markings_ids:
+            if marking_definition_id in stix_entity["markingDefinitionsIds"]:
                 return True
             else:
                 self.opencti.log(
-                    'info',
-                    'Adding Marking-Definition {' + marking_definition_id + '} to Stix-Entity {' + id + '}'
+                    "info",
+                    "Adding Marking-Definition {"
+                    + marking_definition_id
+                    + "} to Stix-Entity {"
+                    + id
+                    + "}",
                 )
                 query = """
                    mutation StixEntityAddRelation($id: ID!, $input: RelationAddInput) {
@@ -354,18 +470,23 @@ class StixEntity:
                        }
                    }
                 """
-                self.opencti.query(query, {
-                    'id': id,
-                    'input': {
-                        'fromRole': 'so',
-                        'toId': marking_definition_id,
-                        'toRole': 'marking',
-                        'through': 'object_marking_refs'
-                    }
-                })
+                self.opencti.query(
+                    query,
+                    {
+                        "id": id,
+                        "input": {
+                            "fromRole": "so",
+                            "toId": marking_definition_id,
+                            "toRole": "marking",
+                            "through": "object_marking_refs",
+                        },
+                    },
+                )
                 return True
         else:
-            self.opencti.log('error', 'Missing parameters: id and marking_definition_id')
+            self.opencti.log(
+                "error", "Missing parameters: id and marking_definition_id"
+            )
             return False
 
     """
@@ -377,21 +498,34 @@ class StixEntity:
     """
 
     def add_tag(self, **kwargs):
-        id = kwargs.get('id', None)
-        stix_entity = kwargs.get('entity', None)
-        tag_id = kwargs.get('tag_id', None)
+        id = kwargs.get("id", None)
+        tag_id = kwargs.get("tag_id", None)
         if id is not None and tag_id is not None:
+            custom_attributes = """
+                id
+                tags {
+                    edges {
+                        node {
+                            id
+                            tag_type
+                            value
+                            color
+                        }
+                        relation {
+                            id
+                        }
+                    }
+                }
+            """
+            stix_entity = self.read(id=id, customAttributes=custom_attributes)
             if stix_entity is None:
-                stix_entity = self.read(id=id)
-            if stix_entity is None:
-                self.opencti.log('error', 'Cannot add Tag, entity not found')
+                self.opencti.log("error", "Cannot add Tag, entity not found")
                 return False
-            if tag_id in stix_entity['tagsIds']:
+            if tag_id in stix_entity["tagsIds"]:
                 return True
             else:
                 self.opencti.log(
-                    'info',
-                    'Adding Tag {' + tag_id + '} to Stix-Entity {' + id + '}'
+                    "info", "Adding Tag {" + tag_id + "} to Stix-Entity {" + id + "}"
                 )
                 query = """
                    mutation StixEntityAddRelation($id: ID!, $input: RelationAddInput) {
@@ -402,18 +536,21 @@ class StixEntity:
                        }
                    }
                 """
-                self.opencti.query(query, {
-                    'id': id,
-                    'input': {
-                        'fromRole': 'so',
-                        'toId': tag_id,
-                        'toRole': 'tagging',
-                        'through': 'tagged'
-                    }
-                })
+                self.opencti.query(
+                    query,
+                    {
+                        "id": id,
+                        "input": {
+                            "fromRole": "so",
+                            "toId": tag_id,
+                            "toRole": "tagging",
+                            "through": "tagged",
+                        },
+                    },
+                )
                 return True
         else:
-            self.opencti.log('error', 'Missing parameters: id and tag_id')
+            self.opencti.log("error", "Missing parameters: id and tag_id")
             return False
 
     """
@@ -425,27 +562,50 @@ class StixEntity:
     """
 
     def add_external_reference(self, **kwargs):
-        id = kwargs.get('id', None)
-        stix_entity = kwargs.get('entity', None)
-        external_reference_id = kwargs.get('external_reference_id', None)
+        id = kwargs.get("id", None)
+        external_reference_id = kwargs.get("external_reference_id", None)
         if id is not None and external_reference_id is not None:
+            custom_attributes = """
+                id
+                externalReferences {
+                    edges {
+                        node {
+                            id
+                            entity_type
+                            stix_id_key
+                            source_name
+                            description
+                            url
+                            hash
+                            external_id
+                            created
+                            modified
+                        }
+                        relation {
+                            id
+                        }
+                    }
+                }
+            """
+            stix_entity = self.read(id=id, customAttributes=custom_attributes)
             if stix_entity is None:
-                stix_entity = self.read(id=id)
-            if stix_entity is None:
-                self.opencti.log('error', 'Cannot add External-Reference, entity not found')
+                self.opencti.log(
+                    "error", "Cannot add External-Reference, entity not found"
+                )
                 return False
-            external_references_ids = []
-            for external_reference in stix_entity['externalReferences']:
-                external_references_ids.append(external_reference['id'])
-            if external_reference_id in external_references_ids:
+            if external_reference_id in stix_entity["externalReferencesIds"]:
                 return True
             else:
                 self.opencti.log(
-                    'info',
-                    'Adding External-Reference {' + external_reference_id + '} to Stix-Entity {' + id + '}'
+                    "info",
+                    "Adding External-Reference {"
+                    + external_reference_id
+                    + "} to Stix-Entity {"
+                    + id
+                    + "}",
                 )
                 query = """
-                   mutation StixEntityAddRelation($id: ID!, $input: RelationAddInput) {
+                   mutation StixEntityEditRelationAdd($id: ID!, $input: RelationAddInput) {
                        stixEntityEdit(id: $id) {
                             relationAdd(input: $input) {
                                 id
@@ -453,18 +613,23 @@ class StixEntity:
                        }
                    }
                 """
-                self.opencti.query(query, {
-                    'id': id,
-                    'input': {
-                        'fromRole': 'so',
-                        'toId': external_reference_id,
-                        'toRole': 'external_reference',
-                        'through': 'external_references'
-                    }
-                })
+                self.opencti.query(
+                    query,
+                    {
+                        "id": id,
+                        "input": {
+                            "fromRole": "so",
+                            "toId": external_reference_id,
+                            "toRole": "external_reference",
+                            "through": "external_references",
+                        },
+                    },
+                )
                 return True
         else:
-            self.opencti.log('error', 'Missing parameters: id and external_reference_id')
+            self.opencti.log(
+                "error", "Missing parameters: id and external_reference_id"
+            )
             return False
 
     """
@@ -476,25 +641,27 @@ class StixEntity:
     """
 
     def add_kill_chain_phase(self, **kwargs):
-        id = kwargs.get('id', None)
-        stix_entity = kwargs.get('entity', None)
-        kill_chain_phase_id = kwargs.get('kill_chain_phase_id', None)
+        id = kwargs.get("id", None)
+        stix_entity = kwargs.get("entity", None)
+        kill_chain_phase_id = kwargs.get("kill_chain_phase_id", None)
         if id is not None and kill_chain_phase_id is not None:
             if stix_entity is None:
                 stix_entity = self.read(id=id)
             if stix_entity is None:
-                self.opencti.log('error', 'Cannot add External-Reference, entity not found')
+                self.opencti.log(
+                    "error", "Cannot add Kill-Chain-Phase, entity not found"
+                )
                 return False
-            stix_entity = self.read(id=id)
-            kill_chain_phases_ids = []
-            for kill_chain_phase in stix_entity['killChainPhases']:
-                kill_chain_phases_ids.append(kill_chain_phase['id'])
-            if kill_chain_phase_id in kill_chain_phases_ids:
+            if kill_chain_phase_id in stix_entity["killChainPhasesIds"]:
                 return True
             else:
                 self.opencti.log(
-                    'info',
-                    'Adding Kill-Chain-Phase {' + kill_chain_phase_id + '} to Stix-Entity {' + id + '}'
+                    "info",
+                    "Adding Kill-Chain-Phase {"
+                    + kill_chain_phase_id
+                    + "} to Stix-Entity {"
+                    + id
+                    + "}",
                 )
                 query = """
                    mutation StixEntityAddRelation($id: ID!, $input: RelationAddInput) {
@@ -505,18 +672,21 @@ class StixEntity:
                        }
                    }
                 """
-                self.opencti.query(query, {
-                    'id': id,
-                    'input': {
-                        'fromRole': 'phase_belonging',
-                        'toId': kill_chain_phase_id,
-                        'toRole': 'kill_chain_phase',
-                        'through': 'kill_chain_phases'
-                    }
-                })
+                self.opencti.query(
+                    query,
+                    {
+                        "id": id,
+                        "input": {
+                            "fromRole": "phase_belonging",
+                            "toId": kill_chain_phase_id,
+                            "toRole": "kill_chain_phase",
+                            "through": "kill_chain_phases",
+                        },
+                    },
+                )
                 return True
         else:
-            self.opencti.log('error', 'Missing parameters: id and kill_chain_phase_id')
+            self.opencti.log("error", "Missing parameters: id and kill_chain_phase_id")
             return False
 
     """
@@ -527,9 +697,9 @@ class StixEntity:
     """
 
     def reports(self, **kwargs):
-        id = kwargs.get('id', None)
+        id = kwargs.get("id", None)
         if id is not None:
-            self.opencti.log('info', 'Getting reports of the Stix-Entity {' + id + '}.')
+            self.opencti.log("info", "Getting reports of the Stix-Entity {" + id + "}.")
             query = """
                 query StixEntity($id: String!) {
                     stixEntity(id: $id) {
@@ -636,6 +806,7 @@ class StixEntity:
                                                 id
                                                 stix_id_key
                                                 entity_type
+                                                observable_value
                                             }
                                             relation {
                                                 id
@@ -662,9 +833,167 @@ class StixEntity:
                     }
                 }
              """
-            result = self.opencti.query(query, {'id': id})
-            processed_result = self.opencti.process_multiple_fields(result['data']['stixEntity'])
-            return processed_result['reports']
+            result = self.opencti.query(query, {"id": id})
+            processed_result = self.opencti.process_multiple_fields(
+                result["data"]["stixEntity"]
+            )
+            if processed_result:
+                return processed_result["reports"]
+            else:
+                return []
         else:
-            self.opencti.log('error', 'Missing parameters: id')
+            self.opencti.log("error", "Missing parameters: id")
+            return None
+
+    """
+        Get the notes about a Stix-Entity object
+
+        :param id: the id of the Stix-Entity
+        :return Stix-Entity object
+    """
+
+    def notes(self, **kwargs):
+        id = kwargs.get("id", None)
+        if id is not None:
+            self.opencti.log("info", "Getting notes of the Stix-Entity {" + id + "}.")
+            query = """
+                query StixEntity($id: String!) {
+                    stixEntity(id: $id) {
+                        notes {
+                            edges {
+                                node {
+                                    id
+                                    stix_id_key
+                                    entity_type
+                                    stix_label
+                                    name
+                                    alias
+                                    description
+                                    content
+                                    graph_data
+                                    created
+                                    modified
+                                    created_at
+                                    updated_at
+                                    createdByRef {
+                                        node {
+                                            id
+                                            entity_type
+                                            stix_id_key
+                                            stix_label
+                                            name
+                                            alias
+                                            description
+                                            created
+                                            modified
+                                        }
+                                        relation {
+                                            id
+                                        }
+                                    }
+                                    markingDefinitions {
+                                        edges {
+                                            node {
+                                                id
+                                                entity_type
+                                                stix_id_key
+                                                definition_type
+                                                definition
+                                                level
+                                                color
+                                                created
+                                                modified
+                                            }
+                                            relation {
+                                                id
+                                            }
+                                        }
+                                    }
+                                    tags {
+                                        edges {
+                                            node {
+                                                id
+                                                tag_type
+                                                value
+                                                color
+                                            }
+                                            relation {
+                                                id
+                                            }
+                                        }
+                                    }            
+                                    externalReferences {
+                                        edges {
+                                            node {
+                                                id
+                                                entity_type
+                                                stix_id_key
+                                                source_name
+                                                description
+                                                url
+                                                hash
+                                                external_id
+                                                created
+                                                modified
+                                            }
+                                            relation {
+                                                id
+                                            }
+                                        }
+                                    }
+                                    objectRefs {
+                                        edges {
+                                            node {
+                                                id
+                                                stix_id_key
+                                                entity_type
+                                            }
+                                            relation {
+                                                id
+                                            }
+                                        }
+                                    }
+                                    observableRefs {
+                                        edges {
+                                            node {
+                                                id
+                                                stix_id_key
+                                                entity_type
+                                                observable_value
+                                            }
+                                            relation {
+                                                id
+                                            }
+                                        }
+                                    }
+                                    relationRefs {
+                                        edges {
+                                            node {
+                                                id
+                                                stix_id_key
+                                            }
+                                            relation {
+                                                id
+                                            }
+                                        }
+                                    }
+                                }
+                                relation {
+                                    id
+                                }
+                            }
+                        }
+                    }
+                }
+             """
+            result = self.opencti.query(query, {"id": id})
+            processed_result = self.opencti.process_multiple_fields(
+                result["data"]["stixEntity"]
+            )
+            if processed_result:
+                return processed_result["notes"]
+            else:
+                return []
+        else:
+            self.opencti.log("error", "Missing parameters: id")
             return None
